@@ -7,31 +7,161 @@
 //!
 //! # Slots and Sinks
 //!
-//! The system of slots and sinks can be relatively hard to understand.
-//! The basic idea is that when a type should be deserialized a "slot"
-//! is passed to it.  The slot is just an `Option` where the final value
-//! is placed.  The system that places these values there is called a
-//! [`Sink`].
+//! Deserialization is based on "slots" and "sinks".  The basic idea is that when a
+//! type should be deserialized a slot in the form of an `Option<T>` is passed
+//! to it where the deserialized value will be placed.  The abstraction that
+//! places these values there is called a [`Sink`] which is returned within a
+//! [`SinkHandle`] from the deserializer.
 //!
-//! While sinks can be implemented on arbitrary types it's more typical to
-//! implement them on a [`SlotWrapper`].  Due to Rust's orphan rules you need
-//! to create your own [`SlotWrapper`] type in your crate by using the
+//! If you can get away with stateless deserialization you can avoid an
+//! allocation by using a newtype wrapper around `Option<T>`.  You can
+//! get such a wrapper by using the
 //! [`make_slot_wrapper`](crate::make_slot_wrapper`) macro ([more
+//! information](https://doc.rust-lang.org/error-index.html#E0117))
+//! which will create a type [`SlotWrapper`].  Due to Rust's orphan rules
+//! you need to create your own type in your crate and you can't use the
+//! one from this module directly.  ([more
 //! information](https://doc.rust-lang.org/error-index.html#E0117)).
 //!
-//! A [`SlotWrapper`] acts as a newtype around an `Option<T>` and derefs into an
-//! `Option<T>`.  To use it implement your desired [`Sink`] for it.  This has the
-//! advantage that it does not need to be allocated.  By calling
-//! [`SlotWrapper::make_handle`] on a slot, one can directly retrieve a
-//! [`SinkHandle`] without the need to box up the slot.
+//! This [`SlotWrapper`] derefs into an `Option<T>` which makes it quite
+//! convenient to use.  By calling [`SlotWrapper::make_handle`] with a slot, one
+//! can directly retrieve a [`SinkHandle`].
+//!
+//! # Deserializing primitives
+//!
+//! To deserialize a primitive you implement a sink for your slot wrapper and
+//! implement the necessary callback.  You can do this as you do not need any
+//! state on the sink so we can use a [`SlotWrapper`].  In this example we
+//! want to accept a `bool` so we just need to implement the
+//! [`atom`](Sink::atom) method as bools are represented as [`Atom`]s.  The
+//! resulting value then must be placed in the slot:
+//!
+//! ```rust
+//! use deser::de::{Sink, Deserialize, DeserializerState, SinkHandle};
+//! use deser::{make_slot_wrapper, Error, Atom};
+//!
+//! make_slot_wrapper!(SlotWrapper);
+//!
+//! struct MyBool(bool);
+//!
+//! impl Sink for SlotWrapper<MyBool> {
+//!     fn atom(
+//!         &mut self,
+//!         atom: Atom,
+//!         state: &DeserializerState,
+//!     ) -> Result<(), Error> {
+//!         match atom {
+//!             Atom::Bool(value) => {
+//!                 // note the extra star here to reach through the deref
+//!                 // of the slot wrapper.
+//!                 **self = Some(MyBool(value));
+//!                 Ok(())
+//!             }
+//!             // for any other value we dispatch to the default handling
+//!             // which creates an unexpected type error but might have
+//!             // more elaborate default behavior in the future.
+//!             other => self.unexpected_atom(other, state)
+//!         }
+//!     }
+//! }
+//!
+//! impl Deserialize for MyBool {
+//!     fn deserialize_into(out: &mut Option<Self>) -> SinkHandle {
+//!         // Since we're using the SlotWrapper abstraction we can directly
+//!         // make a handle here by using the `make_handle` utility.
+//!         SlotWrapper::make_handle(out)
+//!     }
+//! }
+//! ```
+//!
+//! # Struct deserialization
+//!
+//! If you want to deserialize a struct you need to implement the map methods.
+//! As you need to keep track of state you will need to return a boxed sink
+//! and you can't use the slot wrapper.
+//!
+//! ```rust
+//! use deser::de::{DeserializerState, Deserialize, Sink, SinkHandle};
+//! use deser::{Error, ErrorKind};
+//!
+//! struct Flag {
+//!     enabled: bool,
+//!     name: String,
+//! }
+//!
+//! impl Deserialize for Flag {
+//!     fn deserialize_into(out: &mut Option<Self>) -> SinkHandle {
+//!         SinkHandle::boxed(FlagSink {
+//!             out,
+//!             key: None,
+//!             enabled_field: None,
+//!             name_field: None,
+//!         })
+//!     }
+//! }
+//!
+//! struct FlagSink<'a> {
+//!     out: &'a mut Option<Flag>,
+//!     key: Option<String>,
+//!     enabled_field: Option<bool>,
+//!     name_field: Option<String>,
+//! }
+//!     
+//! impl<'a> Sink for FlagSink<'a> {
+//!     fn map(&mut self, _state: &DeserializerState) -> Result<(), Error> {
+//!         // the default implementation returns an error, so we need to
+//!         // override it to remove this error.
+//!         Ok(())
+//!     }
+//!
+//!     fn next_key(&mut self, _state: &DeserializerState) -> Result<SinkHandle, Error> {
+//!         // directly attach to the key field which can hold any
+//!         // string value.  This means that any string is accepted
+//!         // as key.
+//!         Ok(Deserialize::deserialize_into(&mut self.key))
+//!     }
+//!     
+//!     fn next_value(&mut self, state: &DeserializerState) -> Result<SinkHandle, Error> {
+//!         let key = self.key.take().unwrap();
+//!         // since we implement a sink for a struct, move the actual logic for
+//!         // matching into `value_for_key` so that our deserializer can support
+//!         // struct flattening.  If we don't know the key, just return a null
+//!         // handle to ignore it.
+//!         Ok(self.value_for_key(&key, state)?.unwrap_or_else(SinkHandle::null))
+//!     }
+//!
+//!     fn value_for_key(&mut self, key: &str, _state: &DeserializerState)
+//!         -> Result<Option<SinkHandle>, Error>
+//!     {
+//!         Ok(Some(match key {
+//!             "enabled" => Deserialize::deserialize_into(&mut self.enabled_field),
+//!             "name" => Deserialize::deserialize_into(&mut self.name_field),
+//!             _ => return Ok(None)
+//!         }))
+//!     }
+//!     
+//!     fn finish(&mut self, _state: &DeserializerState) -> Result<(), Error> {
+//!         // when we're done, write the final value into the output slot.
+//!         *self.out = Some(Flag {
+//!             enabled: self.enabled_field.take().ok_or_else(|| {
+//!                 Error::new(ErrorKind::MissingField, "field 'enabled' missing")
+//!             })?,
+//!             name: self.name_field.take().ok_or_else(|| {
+//!                 Error::new(ErrorKind::MissingField, "field 'name' missing")
+//!             })?,
+//!         });
+//!         Ok(())
+//!     }
+//! }
+//! ```
 //!
 //! # Driver
 //!
-//! Because the serialization interface of `deser` is tricky to use with lifetimes
-//! without using a lot of stack space, a safe abstraction is provided with the
-//! [`Driver`] type which allow you to drive the deserialization process without
-//! using stack space.  You feed it events and internally the driver ensures
-//! that the deserlization system is driven in the right way.
+//! Because the serialization interface of `deser` is tricky to use because of
+//! lifetimes, a safe abstraction is provided with the [`Driver`] type which
+//! allow you to drive the deserialization process without using stack space.
+//! You feed it events and internally the driver ensures that the deserlization
+//! system is driven in the right way.
 //!
 //! ```rust
 //! use std::collections::BTreeMap;
@@ -54,133 +184,10 @@
 //! assert_eq!(map[&1], "Hello");
 //! assert_eq!(map[&2], "World");
 //! ```
-//!
-//! # Deserializing primitives
-//!
-//! To deserialize a primitive you implement a sink for your slot wrapper and
-//! implement the necessary callback.  For instance to accept a `bool` implement
-//! the [`atom`](Sink::atom) method as bools are represented as [`Atom`]s.  The
-//! resulting value then must be placed in the slot:
-//!
-//! ```rust
-//! use deser::de::{Sink, Deserialize, DeserializerState, SinkHandle};
-//! use deser::{make_slot_wrapper, Error, Atom};
-//!
-//! make_slot_wrapper!(SlotWrapper);
-//!
-//! struct MyBool(bool);
-//!
-//! impl Sink for SlotWrapper<MyBool> {
-//!     fn atom(
-//!         &mut self,
-//!         atom: Atom,
-//!         _state: &DeserializerState,
-//!     ) -> Result<(), Error> {
-//!         match atom {
-//!             Atom::Bool(value) => {
-//!                 // note the extra star here to reach through the deref
-//!                 // of the slot wrapper.
-//!                 **self = Some(MyBool(value));
-//!                 Ok(())
-//!             }
-//!             // for any other value we create an unexpected error.  This is
-//!             // the default implementation of this method.
-//!             other => Err(other.unexpected_error(&self.expecting())),
-//!         }
-//!     }
-//! }
-//!
-//! impl Deserialize for MyBool {
-//!     fn deserialize_into(out: &mut Option<Self>) -> SinkHandle {
-//!         // Since we're using the SlotWrapper abstraction we can directly
-//!         // make a handle here by using the `make_handle` utility.
-//!         SlotWrapper::make_handle(out)
-//!     }
-//! }
-//! ```
-//!
-//! # Struct deserialization
-//!
-//! If you want to deserialize a struct you need to implement a
-//! [`MapSink`] and return it from the main [`Sink`]:
-//!
-//! ```rust
-//! use deser::de::{DeserializerState, Deserialize, Sink, SinkHandle, MapSink};
-//! use deser::{make_slot_wrapper, Error, ErrorKind};
-//!
-//! make_slot_wrapper!(SlotWrapper);
-//!
-//! struct Flag {
-//!     enabled: bool,
-//!     name: String,
-//! }
-//!
-//! impl Deserialize for Flag {
-//!     fn deserialize_into(out: &mut Option<Self>) -> SinkHandle {
-//!         SlotWrapper::make_handle(out)
-//!     }
-//! }
-//!
-//! impl Sink for SlotWrapper<Flag> {
-//!     fn map(
-//!         &mut self,
-//!         _state: &DeserializerState,
-//!     ) -> Result<Box<dyn MapSink + '_>, Error> {
-//!         // return a new map sink for our struct
-//!         Ok(Box::new(FlagMapSink {
-//!             // note that we can directly connect our slot wrapper
-//!             // to the output slot on the sink as it deref's into an Option
-//!             out: self,
-//!             key: None,
-//!             enabled_field: None,
-//!             name_field: None,
-//!         }))
-//!     }
-//! }
-//!     
-//! struct FlagMapSink<'a> {
-//!     out: &'a mut Option<Flag>,
-//!     key: Option<String>,
-//!     enabled_field: Option<bool>,
-//!     name_field: Option<String>,
-//! }
-//!     
-//! impl<'a> MapSink for FlagMapSink<'a> {
-//!     fn key(&mut self) -> Result<SinkHandle, Error> {
-//!         // directly attach to the key field which can hold any
-//!         // string value.  This means that any string is accepted
-//!         // as key.
-//!         Ok(Deserialize::deserialize_into(&mut self.key))
-//!     }
-//!     
-//!     fn value(&mut self) -> Result<SinkHandle, Error> {
-//!         // whenever we are looking for a value slot, look at the last key
-//!         // to decide which value slot to connect.
-//!         match self.key.take().as_deref() {
-//!             Some("enabled") => Ok(Deserialize::deserialize_into(&mut self.enabled_field)),
-//!             Some("name") => Ok(Deserialize::deserialize_into(&mut self.name_field)),
-//!             // if we don't know the key, return a null handle to drop the value.
-//!             _ => Ok(SinkHandle::null()),
-//!         }
-//!     }
-//!     
-//!     fn finish(&mut self, _state: &DeserializerState) -> Result<(), Error> {
-//!         // when we're done, write the final value into the output slot.
-//!         *self.out = Some(Flag {
-//!             enabled: self.enabled_field.take().ok_or_else(|| {
-//!                 Error::new(ErrorKind::MissingField, "field 'enabled' missing")
-//!             })?,
-//!             name: self.name_field.take().ok_or_else(|| {
-//!                 Error::new(ErrorKind::MissingField, "field 'name' missing")
-//!             })?,
-//!         });
-//!         Ok(())
-//!     }
-//! }
-//! ```
 use std::borrow::Cow;
 use std::cell::{Ref, RefMut};
 use std::fmt;
+use std::marker::PhantomData;
 use std::mem::ManuallyDrop;
 use std::ops::{Deref, DerefMut};
 
@@ -191,6 +198,9 @@ use crate::extensions::Extensions;
 
 mod ignore;
 mod impls;
+mod owned;
+
+pub use self::owned::OwnedSink;
 
 __make_slot_wrapper!((pub), SlotWrapper);
 
@@ -296,15 +306,26 @@ pub trait Deserialize: Sized {
 pub trait Sink {
     /// Receives an [`Atom`].
     ///
-    /// The default implementation returns an error.
-    fn atom(&mut self, atom: Atom, _state: &DeserializerState) -> Result<(), Error> {
+    /// Any unknown atom variant should be dispatched to [`unexpected_atom`](Self::unexpected_atom).
+    fn atom(&mut self, atom: Atom, state: &DeserializerState) -> Result<(), Error> {
+        self.unexpected_atom(atom, state)
+    }
+
+    /// Implements a default fallback handling for atoms.
+    fn unexpected_atom(&mut self, atom: Atom, state: &DeserializerState) -> Result<(), Error> {
+        let _ = state;
         Err(atom.unexpected_error(&self.expecting()))
     }
 
-    /// Begins the receiving process for maps.
+    /// Begins the deserialization of a map.
+    ///
+    /// While the deserialization of a map is ongoing the methods
+    /// [`next_key`](Self::next_key) and [`next_value`](Self::next_value) are
+    /// called alternatingly.  The map is ended by [`finish`](Self::finish).
     ///
     /// The default implementation returns an error.
-    fn map(&mut self, _state: &DeserializerState) -> Result<Box<dyn MapSink + '_>, Error> {
+    fn map(&mut self, state: &DeserializerState) -> Result<(), Error> {
+        let _ = state;
         Err(Error::new(
             ErrorKind::Unexpected,
             format!("unexpected map, expected {}", self.expecting()),
@@ -313,58 +334,73 @@ pub trait Sink {
 
     /// Begins the receiving process for sequences.
     ///
+    /// While the deserialization of a sequence is ongoing the method
+    /// [`next_value`](Self::next_value) is called for every new item.
+    /// The sequence is ended by [`finish`](Self::finish).
+    ///
     /// The default implementation returns an error.
-    fn seq(&mut self, _state: &DeserializerState) -> Result<Box<dyn SeqSink + '_>, Error> {
+    fn seq(&mut self, state: &DeserializerState) -> Result<(), Error> {
+        let _ = state;
         Err(Error::new(
             ErrorKind::Unexpected,
             format!("unexpected sequence, expected {}", self.expecting()),
         ))
     }
 
-    /// Utility method to return an expectation message that is used in error messages.
-    fn expecting(&self) -> Cow<'_, str> {
-        "compatible type".into()
-    }
-}
-
-/// A trait to produce sinks for key and value pairs of a map or structs.
-pub trait MapSink {
-    /// Returns the [`Descriptor`] for this map.
-    fn descriptor(&self) -> &dyn Descriptor {
-        &NullDescriptor
+    /// Returns a sink for the next key in a map.
+    fn next_key(&mut self, state: &DeserializerState) -> Result<SinkHandle, Error> {
+        let _ = state;
+        Ok(SinkHandle::null())
     }
 
-    /// Produces the [`Sink`] for the next key.
-    fn key(&mut self) -> Result<SinkHandle, Error>;
+    /// Returns a sink for the next value in a map or sequence.
+    fn next_value(&mut self, state: &DeserializerState) -> Result<SinkHandle, Error> {
+        let _ = state;
+        Ok(SinkHandle::null())
+    }
 
-    /// Produces the [`Sink`] for the next value.
+    /// Returns a value sink for a specific struct field.
     ///
-    /// This can inspect the last key to make a decision about which
-    /// sink to produce.
-    fn value(&mut self) -> Result<SinkHandle, Error>;
+    /// This is a special method that is supposed to be implemented by structs
+    /// if they want to support flattening.  A struct that gets flattened into
+    /// another struct will have this method called to figure out if a key is
+    /// used by it.  The default implementation always returns `None`.
+    fn value_for_key(
+        &mut self,
+        key: &str,
+        state: &DeserializerState,
+    ) -> Result<Option<SinkHandle>, Error> {
+        let _ = key;
+        let _ = state;
+        Ok(None)
+    }
 
-    /// Called when all pairs were produced.
-    fn finish(&mut self, _state: &DeserializerState) -> Result<(), Error>;
-}
+    /// Called after [`atom`](Self::atom), [`map`](Self::map) or [`seq](Self::seq).
+    ///
+    /// The default implementation does nothing.
+    fn finish(&mut self, state: &DeserializerState) -> Result<(), Error> {
+        let _ = state;
+        Ok(())
+    }
 
-/// A trait to produce sinks for items in a sequence.
-pub trait SeqSink {
-    /// Returns the [`Descriptor`] for this seq.
+    /// Returns a descriptor for this type.
     fn descriptor(&self) -> &dyn Descriptor {
         &NullDescriptor
     }
 
-    /// Produces the [`Sink`] for the next item.
-    fn item(&mut self) -> Result<SinkHandle, Error>;
-
-    /// Called when all items were produced.
-    fn finish(&mut self, _state: &DeserializerState) -> Result<(), Error>;
+    /// Utility method to return an expectation message that is used in error messages.
+    ///
+    /// The default implementation returns the type name of the descriptor if available.
+    fn expecting(&self) -> Cow<'_, str> {
+        self.descriptor().name().unwrap_or("compatible type").into()
+    }
 }
 
 /// Gives access to the deserializer state.
 pub struct DeserializerState<'a> {
     extensions: Extensions,
-    stack: ManuallyDrop<Vec<(SinkHandleWrapper, Layer<'a>)>>,
+    descriptor_stack: Vec<&'a dyn Descriptor>,
+    _marker: PhantomData<&'a Driver<'a>>,
 }
 
 impl<'a> DeserializerState<'a> {
@@ -380,17 +416,14 @@ impl<'a> DeserializerState<'a> {
 
     /// Returns the current recursion depth.
     pub fn depth(&self) -> usize {
-        self.stack.len()
+        self.descriptor_stack.len()
     }
 
     /// Returns the topmost descriptor.
     ///
     /// This descriptor always points to a container as the descriptor.
     pub fn top_descriptor(&self) -> Option<&dyn Descriptor> {
-        self.stack.last().map(|x| match &x.1 {
-            Layer::Map(map, _) => map.descriptor(),
-            Layer::Seq(seq) => seq.descriptor(),
-        })
+        self.descriptor_stack.last().copied()
     }
 }
 
@@ -401,8 +434,9 @@ impl<'a> DeserializerState<'a> {
 /// internally impossible with safe code, this is a safe abstractiont that
 /// hides the unsafety internally.
 pub struct Driver<'a> {
-    state: ManuallyDrop<DeserializerState<'a>>,
+    state: DeserializerState<'a>,
     current_sink: Option<SinkHandleWrapper>,
+    sink_stack: ManuallyDrop<Vec<(SinkHandleWrapper, Layer)>>,
 }
 
 struct SinkHandleWrapper {
@@ -419,9 +453,9 @@ impl SinkHandleWrapper {
     }
 }
 
-enum Layer<'a> {
-    Map(Box<dyn MapSink + 'a>, bool),
-    Seq(Box<dyn SeqSink + 'a>),
+enum Layer {
+    Map(bool),
+    Seq,
 }
 
 impl<'a> Driver<'a> {
@@ -433,10 +467,12 @@ impl<'a> Driver<'a> {
     /// Creates a new deserializer driver from a sink.
     pub fn from_sink(sink: SinkHandle) -> Driver<'a> {
         Driver {
-            state: ManuallyDrop::new(DeserializerState {
+            state: DeserializerState {
                 extensions: Extensions::default(),
-                stack: ManuallyDrop::new(Vec::new()),
-            }),
+                descriptor_stack: Vec::new(),
+                _marker: PhantomData,
+            },
+            sink_stack: ManuallyDrop::new(Default::default()),
             current_sink: Some(unsafe { SinkHandleWrapper::from(sink) }),
         }
     }
@@ -462,19 +498,20 @@ impl<'a> Driver<'a> {
     fn _emit(&mut self, event: Event) -> Result<(), Error> {
         macro_rules! target_sink {
             () => {{
-                match self.state.stack.last_mut() {
-                    Some((_, Layer::Map(ref mut map_sink, ref mut is_key))) => {
+                match self.sink_stack.last_mut() {
+                    Some((map_sink, Layer::Map(ref mut is_key))) => {
                         let next_sink = if *is_key {
-                            map_sink.key()?
+                            map_sink.sink.next_key(&self.state)?
                         } else {
-                            map_sink.value()?
+                            map_sink.sink.next_value(&self.state)?
                         };
                         *is_key = !*is_key;
                         self.current_sink = Some(unsafe { SinkHandleWrapper::from(next_sink) });
                     }
-                    Some((_, Layer::Seq(ref mut seq_sink))) => {
-                        self.current_sink =
-                            Some(unsafe { SinkHandleWrapper::from(seq_sink.item()?) });
+                    Some((seq_sink, Layer::Seq)) => {
+                        self.current_sink = Some(unsafe {
+                            SinkHandleWrapper::from(seq_sink.sink.next_value(&self.state)?)
+                        });
                     }
                     _ => {}
                 }
@@ -488,36 +525,46 @@ impl<'a> Driver<'a> {
         }
 
         match event {
-            Event::Atom(atom) => target_sink!().atom(atom, &self.state)?,
+            Event::Atom(atom) => {
+                let current_sink = target_sink!();
+                current_sink.atom(atom, &self.state)?;
+                current_sink.finish(&self.state)?;
+            }
             Event::MapStart => {
                 let current_sink = target_sink!();
-                let map_sink = current_sink.map(&self.state)?;
-                let layer = unsafe { extend_lifetime!(Layer::Map(map_sink, true), Layer<'_>) };
+                current_sink.map(&self.state)?;
+                let descriptor = current_sink.descriptor();
                 self.state
-                    .stack
-                    .push((self.current_sink.take().unwrap(), layer));
+                    .descriptor_stack
+                    .push(unsafe { extend_lifetime!(descriptor, &dyn Descriptor) });
+                self.sink_stack
+                    .push((self.current_sink.take().unwrap(), Layer::Map(true)));
                 return Ok(());
             }
-            Event::MapEnd => match self.state.stack.pop() {
-                Some((next_sink, Layer::Map(mut map_sink, _))) => {
-                    map_sink.finish(&self.state)?;
-                    self.current_sink = Some(next_sink);
+            Event::MapEnd => match self.sink_stack.pop() {
+                Some((mut map_sink, Layer::Map(_))) => {
+                    map_sink.sink.finish(&self.state)?;
+                    self.state.descriptor_stack.pop();
+                    self.current_sink = Some(map_sink);
                 }
                 _ => panic!("not inside a MapSink"),
             },
             Event::SeqStart => {
                 let current_sink = target_sink!();
-                let seq_sink = current_sink.seq(&self.state)?;
-                let layer = unsafe { extend_lifetime!(Layer::Seq(seq_sink), Layer<'_>) };
+                current_sink.seq(&self.state)?;
+                let descriptor = current_sink.descriptor();
                 self.state
-                    .stack
-                    .push((self.current_sink.take().unwrap(), layer));
+                    .descriptor_stack
+                    .push(unsafe { extend_lifetime!(descriptor, &dyn Descriptor) });
+                self.sink_stack
+                    .push((self.current_sink.take().unwrap(), Layer::Seq));
                 return Ok(());
             }
-            Event::SeqEnd => match self.state.stack.pop() {
-                Some((next_sink, Layer::Seq(mut seq_sink))) => {
-                    seq_sink.finish(&self.state)?;
-                    self.current_sink = Some(next_sink);
+            Event::SeqEnd => match self.sink_stack.pop() {
+                Some((mut seq_sink, Layer::Seq)) => {
+                    seq_sink.sink.finish(&self.state)?;
+                    self.state.descriptor_stack.pop();
+                    self.current_sink = Some(seq_sink);
                 }
                 _ => panic!("not inside a SeqSink"),
             },
@@ -531,13 +578,11 @@ impl<'a> Driver<'a> {
 
 impl<'a> Drop for Driver<'a> {
     fn drop(&mut self) {
-        // it's important that we drop the values in inverse order.
-        while let Some(_last) = self.state.stack.pop() {
-            // drop in inverse order
-        }
         unsafe {
-            ManuallyDrop::drop(&mut self.state.stack);
-            ManuallyDrop::drop(&mut self.state);
+            while let Some(_item) = self.sink_stack.pop() {
+                // drop in inverse order
+            }
+            ManuallyDrop::drop(&mut self.sink_stack);
         }
     }
 }
